'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _propTypes = require('prop-types');

var _propTypes2 = _interopRequireDefault(_propTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var StickyTree = function (_React$PureComponent) {
    _inherits(StickyTree, _React$PureComponent);

    function StickyTree(props) {
        _classCallCheck(this, StickyTree);

        var _this = _possibleConstructorReturn(this, (StickyTree.__proto__ || Object.getPrototypeOf(StickyTree)).call(this, props));

        _this.onScroll = _this.onScroll.bind(_this);

        _this.state = {
            currNodePos: 0
        };

        _this.scrollTop = 0;
        _this.nodePosCache = [];
        _this.rowRenderRange = undefined;
        return _this;
    }

    /**
     *  Converts the consumer's tree structure into a flat array with root at index: 0,
     *  including information about the top and height of each node.
     *
     *  i.e:
     *  [
     *    { node: 'root', top: 0, index: 0, height: 100 },
     *    { node: 'child1', top: 10, index: 0, parentIndex: 0 height: 10 },
     *    ...
     *  ]
     */


    _createClass(StickyTree, [{
        key: 'flattenTree',
        value: function () {
            function flattenTree(node) {
                var nodes = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
                var context = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : { totalHeight: 0, parentIndex: undefined };

                var index = nodes.length;
                var nodeInfo = { node: node, top: context.totalHeight, parentIndex: context.parentIndex, index: index };
                nodes.push(nodeInfo);

                if (context.parentIndex !== undefined) {
                    nodes[context.parentIndex].children.push(index);
                }

                context.totalHeight += this.props.getHeight(node);

                var children = this.props.getChildren(node);
                if (Array.isArray(children)) {
                    nodeInfo.children = [];
                    for (var i = 0; i < children.length; i++) {
                        // Need to reset parentIndex here as we are recursive.
                        context.parentIndex = index;
                        var child = children[i];
                        this.flattenTree(child, nodes, context);
                    }
                }
                nodeInfo.height = context.totalHeight - nodeInfo.top;

                return nodes;
            }

            return flattenTree;
        }()
    }, {
        key: 'componentWillMount',
        value: function () {
            function componentWillMount() {
                if (this.props.root !== undefined) {
                    this.nodePosCache = this.flattenTree(this.props.root);
                }
            }

            return componentWillMount;
        }()
    }, {
        key: 'componentWillReceiveProps',
        value: function () {
            function componentWillReceiveProps(newProps) {
                if (newProps.root !== this.props.root) {
                    this.nodePosCache = this.flattenTree(newProps.root);
                }
                if (newProps.scrollTop !== this.scrollTop) {
                    this.elem.scrollTop = newProps.scrollTop;
                }
            }

            return componentWillReceiveProps;
        }()
    }, {
        key: 'componentDidUpdate',
        value: function () {
            function componentDidUpdate(prevProps, prevState) {
                if (this.props.onRowsRendered !== undefined && prevState.currNodePos !== this.props.currNodePos) {
                    var range = this.rowRenderRange;

                    this.props.onRowsRendered({
                        overscanStartIndex: range.start,
                        overscanStopIndex: range.end,
                        startIndex: range.visibleStart,
                        stopIndex: range.visibleEnd,
                        startNode: this.nodePosCache[range.visibleStart].node,
                        endNode: this.nodePosCache[range.visibleEnd].node
                    });
                }
            }

            return componentDidUpdate;
        }()
    }, {
        key: 'getChildContainerStyle',
        value: function () {
            function getChildContainerStyle(child, top) {
                return { position: 'absolute', top: top, height: child.height, width: '100%' };
            }

            return getChildContainerStyle;
        }()
    }, {
        key: 'renderParentTree',
        value: function () {
            function renderParentTree() {
                this.rowRenderRange = this.getRenderRowRange();
                var path = this.getParentPath(this.rowRenderRange.start);

                // Parent nodes to the current range.
                var indexesToRender = new Set();
                for (var i = 0; i < path.length; i++) {
                    indexesToRender.add(path[i].index);
                }

                // The rest of the nodes within the range.
                for (var _i = this.rowRenderRange.start; _i <= this.rowRenderRange.end; _i++) {
                    indexesToRender.add(this.nodePosCache[_i].index);
                }

                if (this.props.renderRoot) {
                    return _react2['default'].createElement(
                        'div',
                        { className: 'rv-sticky-node-list', style: { width: '100%', position: 'absolute', top: 0 } },
                        this.renderChildWithChildren(path[0], 0, indexesToRender)
                    );
                }
                return this.renderParentContainer(path[0], indexesToRender);
            }

            return renderParentTree;
        }()
    }, {
        key: 'renderParentContainer',
        value: function () {
            function renderParentContainer(parent, indexesToRender) {
                return _react2['default'].createElement(
                    'div',
                    { key: parent.node, className: 'rv-sticky-node-list', style: { position: 'absolute', width: '100%' } },
                    this.renderChildren(parent, indexesToRender)
                );
            }

            return renderParentContainer;
        }()
    }, {
        key: 'renderChildWithChildren',
        value: function () {
            function renderChildWithChildren(child, top, indexesToRender) {
                return _react2['default'].createElement(
                    'div',
                    { className: 'rv-sticky-parent-node', key: child.node, style: this.getChildContainerStyle(child, top) },
                    this.props.rowRenderer(child.node),
                    this.renderParentContainer(child, indexesToRender)
                );
            }

            return renderChildWithChildren;
        }()
    }, {
        key: 'renderChildren',
        value: function () {
            function renderChildren(parent, indexesToRender) {
                var _this2 = this;

                var nodes = [];
                var top = 0;
                parent.children.forEach(function (index) {
                    var child = _this2.nodePosCache[index];

                    if (indexesToRender.has(index)) {
                        if (child.children) {
                            nodes.push(_this2.renderChildWithChildren(child, top, indexesToRender));
                        } else {
                            nodes.push(_react2['default'].createElement(
                                'div',
                                {
                                    className: 'rv-sticky-leaf-node', key: child.node,
                                    style: _this2.getChildContainerStyle(child, top) },
                                _this2.props.rowRenderer(child.node)
                            ));
                        }
                    }
                    // Needs to be on the outside so that we add the the top even if
                    // this node is not visible
                    top += child.height;
                });
                return nodes;
            }

            return renderChildren;
        }()

        /**
         * Determines the start and end number of the range to be rendered.
         * @returns {{start: number, end: number}} Indexes within nodePosCache
         */

    }, {
        key: 'getRenderRowRange',
        value: function () {
            function getRenderRowRange() {
                var start = this.state.currNodePos - this.props.overscanRowCount;
                if (start < 0) {
                    start = 0;
                }
                var visibleEnd = this.state.currNodePos + 1;

                while (this.nodePosCache[visibleEnd] && this.nodePosCache[visibleEnd].top < this.scrollTop + this.props.height) {
                    visibleEnd++;
                }

                var end = visibleEnd + this.props.overscanRowCount;
                if (end > this.nodePosCache.length - 1) {
                    end = this.nodePosCache.length - 1;
                }

                return { start: start, end: end, visibleStart: this.state.currNodePos, visibleEnd: visibleEnd };
            }

            return getRenderRowRange;
        }()

        /**
         * Returns the parent path for the specified index within nodePosCache.
         * @param nodeIndex
         * @returns {Array<Node>}
         */

    }, {
        key: 'getParentPath',
        value: function () {
            function getParentPath(nodeIndex) {
                var currNode = this.nodePosCache[nodeIndex];
                var path = [currNode];
                while (currNode) {
                    currNode = this.nodePosCache[currNode.parentIndex];
                    if (currNode) {
                        path.push(currNode);
                    }
                }
                return path.reverse();
            }

            return getParentPath;
        }()
    }, {
        key: 'forwardSearch',
        value: function () {
            function forwardSearch(scrollTop) {
                var nodePosCache = this.nodePosCache;
                for (var i = this.state.currNodePos; i < nodePosCache.length; i++) {
                    if (nodePosCache[i].top >= scrollTop) {
                        return i;
                    }
                }
                return nodePosCache.length - 1;
            }

            return forwardSearch;
        }()
    }, {
        key: 'backwardSearch',
        value: function () {
            function backwardSearch(scrollTop) {
                var nodePosCache = this.nodePosCache;
                for (var i = this.state.currNodePos; i >= 0; i--) {
                    if (nodePosCache[i].top <= scrollTop) {
                        return i;
                    }
                }
                return 0;
            }

            return backwardSearch;
        }()

        /**
         * Returns the closest node within nodePosCache.
         * @param scrollTop
         */

    }, {
        key: 'findClosestNode',
        value: function () {
            function findClosestNode(scrollTop) {
                var pos = void 0;
                if (scrollTop > this.scrollTop) {
                    pos = this.forwardSearch(scrollTop);
                } else if (scrollTop < this.scrollTop) {
                    pos = this.backwardSearch(scrollTop);
                }
                if (pos !== this.state.currNodePos) {
                    this.setState({ currNodePos: pos });
                }
            }

            return findClosestNode;
        }()
    }, {
        key: 'onScroll',
        value: function () {
            function onScroll(e) {
                var scrollTop = e.target.scrollTop;
                this.findClosestNode(scrollTop);
                this.scrollTop = scrollTop;

                if (this.props.onScroll !== undefined) {
                    this.props.onScroll({ scrollTop: this.scrollTop });
                }
            }

            return onScroll;
        }()
    }, {
        key: 'getStyle',
        value: function () {
            function getStyle() {
                var style = { overflow: 'auto', position: 'relative' };
                if (this.props.width) {
                    style.width = this.props.width;
                }
                if (this.props.height) {
                    style.height = this.props.height;
                }
                return style;
            }

            return getStyle;
        }()
    }, {
        key: 'render',
        value: function () {
            function render() {
                var _this3 = this;

                return _react2['default'].createElement(
                    'div',
                    { ref: function () {
                            function ref(elem) {
                                return _this3.elem = elem;
                            }

                            return ref;
                        }(), className: 'rv-sticky-tree', style: this.getStyle(), onScroll: this.onScroll },
                    this.renderParentTree()
                );
            }

            return render;
        }()
    }]);

    return StickyTree;
}(_react2['default'].PureComponent);

StickyTree.propTypes = {
    getChildren: _propTypes2['default'].func.isRequired,
    getHeight: _propTypes2['default'].func.isRequired,
    rowRenderer: _propTypes2['default'].func.isRequired,
    root: _propTypes2['default'].any.isRequired,
    overscanRowCount: _propTypes2['default'].number.isRequired,
    height: _propTypes2['default'].number,
    width: _propTypes2['default'].number,
    renderRoot: _propTypes2['default'].bool,

    /**
     * Called whenever the scrollTop position changes.
     */
    onScroll: _propTypes2['default'].func,

    /**
     * Called to indicate that a new render range for rows has been rendered.
     */
    onRowsRendered: _propTypes2['default'].func
};
StickyTree.defaultProps = {
    overscanRowCount: 0,
    renderRoot: true
};
exports['default'] = StickyTree;